import 'dart:async';
import 'package:captain/src/captain_route_parser.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';

/// Configuration for a Captain widget
///
/// provide:
/// - list of initial pages
/// - popPage function where you should implement pop logic
/// - shouldPop function to decide whether to handle pop request
/// - actions that are callable via a dynamic key object and individually modify the page stack and make the navigator rebuild ->
/// return a List<Page> depending on current page stack
class CaptainConfig with ChangeNotifier {
  CaptainConfig(
    /// stack of pages to be initially built
    this._pages, {

    /// function that decides how to pop a page off the page stack
    this.popPage,

    /// function that decides on if the current page that received a pop request should be popped or not (if false the app closes)
    this.shouldPop,

    /// set of actions that can be invoked by Navigator.of().invokeAction('key')
    this.actions,

    /// asynchronously returns a page for a new routeinformation coming in which will be set as a new route path
    this.parseRouteInformation,

    /// the initial routeinformation the app should start with
    this.initialRouteInformation,

    /// how to restore the routeinformation
    this.restoreRouteInformation,
    this.observers,
    this.routeParser,
  });
  List<Page> _pages;

  /// exposes the list pages that are built by the navigator (unmodifiable)
  List<Page> get pages => List.unmodifiable(_pages);

  /// navigator navigatorKey
  final GlobalKey<NavigatorState> navigatorKey = GlobalKey<NavigatorState>();

  // ### POP PAGE ###

  @protected
  final bool? Function(Route<dynamic>, dynamic, List<Page>)? popPage;

  bool popPageRouterFacing(Route<dynamic> _route, dynamic _result) {
    return popPage != null
        ? popPage!(_route, _result, pages)!
        : _defaultPopPage(_route, _result);
  }

  bool _defaultPopPage(Route<dynamic> _route, dynamic _result) {
    bool _res = _route.didPop(_result);
    _res ? _pages.removeLast() : _res = false;
    return _res;
  }

  // ### RÓUTE PARSER ###
  final CaptainRouteParser? routeParser;

  // ### PARSE ROUTE INFORMATION ###

  final Future<Page> Function(
    RouteInformation routeInformation,
    List<Page> pages,
  )? parseRouteInformation;

  Future<Page> parseRouteInformationRouterFacing(
      RouteInformation _routeInformation) async {
    return parseRouteInformation != null
        ? parseRouteInformation!(_routeInformation, pages)
        : _defaultParseRouteInformation(_routeInformation);
  }

  Future<Page> _defaultParseRouteInformation(
      RouteInformation _routeInformation) {
    return SynchronousFuture<Page>(pages.first);
  }

  // ### INITIAL ROUTE INFORMATION ###

  final RouteInformation? initialRouteInformation;

  RouteInformation intialRouteInformationRouterFacing() {
    return initialRouteInformation == null
        ? _defaultInitialRouteInformation()
        : initialRouteInformation!;
  }

  RouteInformation _defaultInitialRouteInformation() {
    return RouteInformation(location: "/");
  }

  // ### RESTORE ROUTE INFORMATION ###

  final RouteInformation Function(Page page)? restoreRouteInformation;

  RouteInformation restoreRouteInformationRouterFacing(Page _page) {
    return restoreRouteInformation != null
        ? restoreRouteInformation!(_page)
        : _defaultRestoreRouteInformation(_page);
  }

  RouteInformation _defaultRestoreRouteInformation(Page _page) {
    return RouteInformation(location: '/');
  }

  // ### SHOULD POP ###

  @protected
  late Future<bool>? Function(List<Page> pages)? shouldPop;

  Future<bool> shouldPopRouterFacing() async {
    return shouldPop == null ? _defaultShouldPop() : shouldPop!(pages)!;
  }

  Future<bool> _defaultShouldPop() {
    return SynchronousFuture<bool>(pages.isNotEmpty);
  }

  // ### OBSERVERS ###

  final List<void Function(List<Page> pages)>? observers;

  void invokeObservers() {
    if (observers == null) {
      return;
    }
    for (final _observer in observers!) {
      _observer(_pages);
    }
  }

  // ### ACTIONS ###

  @protected
  final Map<dynamic, List<Page> Function(List<Page> pages)>? actions;

  void _afterAction() {
    notifyListeners();
    invokeObservers();
  }

  void invokeAction(dynamic actionKey) {
    _pages = actions![actionKey]!(_pages);
    _afterAction();
  }

  void invokeActionFunc(List<Page> Function(List<Page> pages) actionFunc) {
    final _restorePageStack = _pages;
    _pages = actionFunc(_pages);
    if (_pages.isEmpty) {
      _pages = _restorePageStack;
    }
    _afterAction();
  }

  void upsertAction(
      List<Page> Function(List<Page>) actionFunc, dynamic actionKey) {
    if (actions!.containsKey(actionKey)) {
      actions!.update(actionKey, (_) => actionFunc);
      return;
    }
    actions![actionKey] = actionFunc;
    _afterAction();
  }

  void push(Page _page) {
    _pages.add(_page);
    _afterAction();
  }

  void pop() {
    _afterAction();
  }
}

/// ### Captain widget "Router"
///
/// place the Captain widget beneath your MaterialApp and pass it a CaptainConfig
/// - Captain widgets can be nested just like Router
/// - Captain is compatible with Navigator imperative API style of calling
/// ````dart
/// Navigator.of(context).action("myActionKey");
/// Navigator.of(context).actionFunc((pageStack) => pageStack..add(pageToAdd));
/// ````
class Captain extends InheritedWidget {
  Captain({
    required this.config,
  }) : super(
          child: Router(
            routerDelegate: _AppRouterDelegate(
              config,
            ),
            routeInformationParser: _AppRouteInformationParser(
              config,
            ),
            routeInformationProvider: PlatformRouteInformationProvider(
              initialRouteInformation:
                  config.intialRouteInformationRouterFacing(),
            ),
          ),
        );

  final CaptainConfig config;

  @override
  bool updateShouldNotify(covariant Captain oldWidget) {
    return oldWidget.config != config;
  }

  static Captain? of(BuildContext context) {
    return context.dependOnInheritedWidgetOfExactType<Captain>()!;
  }
}

class _AppRouterDelegate extends RouterDelegate<Page>
    with ChangeNotifier, PopNavigatorRouterDelegateMixin {
  _AppRouterDelegate(
    this.config,
  ) {
    config.addListener(notifyListeners);
  }
  final CaptainConfig config;

  @override
  Widget build(BuildContext context) {
    return Navigator(
      key: config.navigatorKey,
      pages: config.pages,
      onPopPage: config.popPageRouterFacing,
    );
  }

  @override
  Future<bool> popRoute() {
    print(config.shouldPopRouterFacing());
    return config.shouldPopRouterFacing();
  }

  @override
  Future<void> setNewRoutePath(Page _page) {
    return SynchronousFuture(() {
      config.push(_page);
    }());
  }

  @override
  GlobalKey<NavigatorState> get navigatorKey => config.navigatorKey;
}

class _AppRouteInformationParser extends RouteInformationParser<Page> {
  _AppRouteInformationParser(this.config);
  final CaptainConfig config;

  @override
  Future<Page> parseRouteInformation(RouteInformation _routeInformation) async {
    if (_routeInformation.location != null || config.routeParser != null) {
      return SynchronousFuture(
          config.routeParser!.parseRouteMapToPage(_routeInformation.location!));
    }
    return await config.parseRouteInformationRouterFacing(_routeInformation);
  }

  @override
  RouteInformation restoreRouteInformation(Page _page) =>
      config.restoreRouteInformationRouterFacing(_page);
}
